open Printf
open Atom
open Types
open Terms
open Coercions
open Print
open Typerr
open Typematch


(* ------------------------------------------------------------------------- *)

(* The core of the typechecker is made up of two mutually recursive
   functions. [infer] infers a type and returns it; [check] infers a
   type, checks that it is equal to the expected type, and returns
   nothing. *)

(* The typechecker ignores the type annotations carried by the term,
   and overwrites them with fresh annotations, which are guaranteed to
   be correct. *)

let rec infer              (* [infer] expects... *)
    (p : Export.env)       (* a pretty-printer environment, for printing types; *)
    (loc : Error.location) (* a location, for reporting errors; *)
    (tenv : tenv)          (* a typing environment; *)
    (term : loc_fterm)     (* a term-with-location; *)
    : ftype * loc_fterm =  (* ...and returns a type. *)

  (* First, determine the new current location. *)

  let loc, term = Error.override loc term in

  (* Then, examine the term at hand. *)

  match term with

       | _ ->
       failwith "TYPECHECKING IS NOT IMPLEMENTED YET!" (* do something here! *)

and check                  (* [check] expects... *)
    (p : Export.env)       (* a pretty-printer environment, for printing types; *)
    (loc : Error.location) (* a location, for reporting errors; *)
    (tenv : tenv)          (* a typing environment; *)
    (term : loc_fterm)     (* a term-with-location; *)
    (expected : ftype)     (* an expected type; *)
    : loc_fterm =          (* ...and returns nothing. *)

  let inferred, term' = infer p loc tenv term in
  if not (equal expected inferred) then
    let loc, _ = Error.override loc term in
    mismatch p loc expected inferred
  else term'

(* ------------------------------------------------------------------------- *)

(* A complete program is typechecked within empty environments. *)

let program (term : loc_fterm) : ftype * loc_fterm =
  let p = Export.empty
  and loc = Error.dummy
  and tenv = Types.empty in
  let res = infer p loc tenv term in
  (* Some errors, such as multiple definitions, have been delayed *)
  Error.signaled();
  res

